--[[
    hyper-ecs, an ECS framework for LÖVE
    Copyright (C) 2018 Evan Hemsley

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--]]

local class = require "middleclass"

local Entity = require('hyper.entity')
local Component = require('hyper.component')
local Engine = require('hyper.engine')

local World = class('World')

local function intersect(set_a, set_b)
    local new_set = {}
    for k, _ in pairs(set_a) do
        new_set[k] = set_b[k]
    end
    return new_set
end

function World:initialize()
    self.next_id_value = 1

    self.engines = {}
    self.entities = {}

    self.component_entity_id_map = {}
end

function World:next_id()
    local id = self.next_id_value
    self.next_id_value = self.next_id_value + 1
    return id
end

function World:create_entity()
    local id = self:next_id()
    local entity = Entity:new(id, self)

    self.entities[id] = entity
    return entity
end

function World:create_component(type, entity)
    local component = Component:new(type, entity)
    if self.component_entity_id_map[type] == nil then
        self.component_entity_id_map[type] = { }
    end

    self.component_entity_id_map[type][entity.id] = entity.id -- break this set logic into a class?
    return component
end

function World:entities_with_components(types)
    local entities_per_component_type = {}
    for _, type in pairs(types) do
        table.insert(entities_per_component_type, self.component_entity_id_map[type])
    end

    local ids = entities_per_component_type[1]
    for _, entities_list in ipairs(entities_per_component_type) do
        ids = intersect(ids, entities_list)
    end

    local result = {}
    for _, id in pairs(ids) do
        table.insert(result, self:find_entity(id))
    end
    return result
end

function World:create_engine(component_types)
    local engine = Engine:new(component_types)
    table.insert(self.engines, engine)
    return engine
end

function World:check_and_register_entity(entity)
    for _, engine in ipairs(self.engines) do
        engine:check_and_register_entity(entity)
    end
end

function World:find_entity(id)
    return self.entities[id]
end

return World