local class = require "middleclass"

local World = require('hyper').World
local Component = require('hyper').Component

local helper = require('spec.helper')
assert:register(
    "assertion",
    "has_value",
    helper.has_value,
    "assertion.has_value.positive",
    "assertion.has_value.negative"
)

describe('World', function ()
    describe('next_id', function ()
        it('should start at one', function ()
            local world = World:new()
            assert.are.equal(world:next_id(), 1)
        end)

        it('should increment by 1 every time called', function ()
            local world = World:new()
            world:next_id()
            world:next_id()
            assert.are.equal(world:next_id(), 3)
        end)
    end)

    describe('create_entity', function()
        it('should assign an id', function()
            local world = World:new()
            world:next_id()

            local entity = world:create_entity()
            assert.are.equal(entity.id, 2)
        end)

        it('should assign world', function()
            local world = World:new()
            local entity = world:create_entity()
            assert.are.equal(world, entity.world)
        end)

        it('should create a reference to the entity by id', function()
            local world = World:new()
            local entity = world:create_entity()

            assert.are.equal(world.entities[entity.id], entity)
        end)
    end)

    describe ('create_component', function()
        local world = World:new()
        local entity = world:create_entity()
        local TestComponent = class('TestComponent', Component) -- shorthand this
        local component = world:create_component(TestComponent, entity)

        it('should assign the entity to the component', function()
            assert.are.equal(component.entity, entity)
        end)

        it('should create a component of given type', function()
            assert.are.equal(component.type, TestComponent)
        end)
    end)

    describe('entities_with_components', function()
        it ('should return all entities with given components', function()
            local world = World:new()
            local A = class('A', Component)
            local B = class('B', Component)
            local C = class('C', Component)
            local entity_ab = world:create_entity()
            local entity_ac = world:create_entity()
            local entity_abc = world:create_entity()

            world:create_component(A, entity_ab)
            world:create_component(B, entity_ab)

            world:create_component(A, entity_ac)
            world:create_component(C, entity_ac)

            world:create_component(A, entity_abc)
            world:create_component(B, entity_abc)
            world:create_component(C, entity_abc)

            local entities_with_ab_components = world:entities_with_components({A,B})

            assert.has_value(entities_with_ab_components, entity_ab)
            assert.has_value(entities_with_ab_components, entity_abc)
            assert.is_not.has_value(entities_with_ab_components, entity_ac)
        end)
    end)

    it('creates engine', function()
        local world = World:new()
        local engine = world:create_engine({})
        assert.has_value(world.engines, engine)
    end)

    it('calls the check and register function on all engines', function()
        local world = World:new()
        local engine_one = world:create_engine()
        local engine_two = world:create_engine()
        local result = 1
        local function engine_check(entity)
            result = result + 1
        end
        engine_one.check_and_register_entity = engine_check
        engine_two.check_and_register_entity = engine_check

        world:check_and_register_entity({})
        assert.are.same(result, 3)
    end)

    it('find_entity', function()
        local world = World:new()
        local entity = world:create_entity()
        assert.are.equal(world:find_entity(entity.id), entity)
    end)
end)